package net.sourceforge.nattable.group;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

import net.sourceforge.nattable.layer.IUniqueIndexLayer;
import net.sourceforge.nattable.persistence.IPersistable;
import net.sourceforge.nattable.util.ObjectUtils;

/**
 * Tracks:
 *  Rows (by index) in a row Group. Does not keep the row indexes in any defined order.
 *  Expand/collapse state of the Group.
 *  Name of the Row Group (RG)
 */
public class RowGroupModel implements IPersistable {

	private static final String PERSISTENCE_KEY_ROW_GROUPS = ".rowGroups";

	/** Row group header name to row indexes */
	private final LinkedHashMap<String, RowGroup> nameToRowGroup = new LinkedHashMap<String, RowGroup>();

	private final LinkedHashMap<Integer, String> indexToRowGroupName = new LinkedHashMap<Integer, String>();

	private final Collection<IRowGroupModelListener> listeners = new HashSet<IRowGroupModelListener>();

	public void registerRowGroupModelListener(IRowGroupModelListener listener) {
		listeners.add(listener);
	}

	public void notifyListeners() {
		for (IRowGroupModelListener listener : listeners) {
			listener.rowGroupModelChanged();
		}
	}
	
	public void saveState(String prefix, Properties properties) {
		StringBuilder strBuilder = new StringBuilder();

		for (String rowGroupName : nameToRowGroup.keySet()) {
			
			RowGroup rowGroup = nameToRowGroup.get(rowGroupName);
			
			// if this rowGroup has members, continue without saving state.
			// A group can haven no members if groups are used to organise 
			// rows on a higher abstraction level ...
			if (rowGroup.members.size() == 0) {
				continue;
			}
			
			strBuilder.append(rowGroupName);
			strBuilder.append('=');

			strBuilder.append(rowGroup.collapsed ? "collapsed" : "expanded");
			strBuilder.append(':');

			strBuilder.append(rowGroup.collapseable ? "collapseable" : "uncollapseable");
			strBuilder.append(':');
			
			strBuilder.append(rowGroup.unbreakable ? "unbreakable" : "breakable");
			strBuilder.append(':');

			for (Integer member : rowGroup.members) {
				strBuilder.append(member);
				strBuilder.append(',');
			}

			strBuilder.append('|');
		}

		properties.setProperty(prefix + PERSISTENCE_KEY_ROW_GROUPS, strBuilder.toString());
	}

	public void loadState(String prefix, Properties properties) {
		String property = properties.getProperty(prefix + PERSISTENCE_KEY_ROW_GROUPS);
		if (property != null) {
			clear();

			StringTokenizer rowGroupTokenizer = new StringTokenizer(property, "|");
			while (rowGroupTokenizer.hasMoreTokens()) {
				String rowGroupToken = rowGroupTokenizer.nextToken();

				int separatorIndex = rowGroupToken.indexOf('=');

				// Row group name
				String rowGroupName = rowGroupToken.substring(0, separatorIndex);
				RowGroup rowGroup = new RowGroup(rowGroupName);
				nameToRowGroup.put(rowGroupName, rowGroup);

				String[] rowGroupProperties = rowGroupToken.substring(separatorIndex + 1).split(":");

				// Expanded/collapsed
				String state = rowGroupProperties[0];
				if ("collapsed".equals(state)) {
					rowGroup.collapsed = true;
				} else if ("expanded".equals(state)) {
					rowGroup.collapsed = false;
				} else {
					throw new IllegalArgumentException(state + " not one of 'expanded' or 'collapsed'");
				}

				// collapseble / uncollapseable
				state = rowGroupProperties[1];
				if ("collapseable".equals(state)) {
					rowGroup.collapseable = true;
				} else if ("uncollapseable".equals(state)) {
					rowGroup.collapseable = false;
				} else {
					throw new IllegalArgumentException(state + " not one of 'uncollapseable' or 'collapseable'");
				}
				
				// breakable / unbreakable
				state = rowGroupProperties[2];
				if ("breakable".equals(state)) {
					rowGroup.unbreakable = false;
				} else if ("unbreakable".equals(state)) {
					rowGroup.unbreakable = true;
				} else {
					throw new IllegalArgumentException(state + " not one of 'breakable' or 'unbreakable'");
				}

				// Indexes
				String indexes = rowGroupProperties[3];
				StringTokenizer indexTokenizer = new StringTokenizer(indexes, ",");
				while (indexTokenizer.hasMoreTokens()) {
					Integer index = Integer.valueOf(indexTokenizer.nextToken());
					rowGroup.members.add(index);
					indexToRowGroupName.put(index, rowGroupName);
				}
			}
		}
	}

	/**
	 * Creates the row group if one does not exist with the given name
	 * and adds the row indexes to it.
	 * @see RowGroupModel#insertRowIndexes(String, int...);
	 */
	public void addRowsIndexesToGroup(String rowGroupName, int... bodyRowIndexes) {
		if (nameToRowGroup.get(rowGroupName) == null) {
			RowGroup group = new RowGroup(rowGroupName);
			nameToRowGroup.put(rowGroupName, group);
		}
		insertRowIndexes(rowGroupName, bodyRowIndexes);
		notifyListeners();
	}
	
	/**
	 * This method will add row index(s) to an existing group
	 * @param rowGroupName to add the indexes to
	 * @param rowIndexesToInsert
	 * @return FALSE if:
	 * 		The row group is frozen
	 * 		Index is already s part of a row group
	 */
	public boolean insertRowIndexes(String rowGroupName, int... rowIndexesToInsert) {
		LinkedList<Integer> members = new LinkedList<Integer>();
		LinkedHashMap<Integer, String> indexToRowGroupName = new LinkedHashMap<Integer, String>();

		RowGroup rowGroup = nameToRowGroup.get(rowGroupName);
		if (rowGroup.unbreakable) {
			return false;
		}

		// Check if any of the indexes belong to existing groups
		for (int rowIndexToInsert : rowIndexesToInsert) {
			final Integer index = Integer.valueOf(rowIndexToInsert);
			if (isPartOfAGroup(rowIndexToInsert)) {
				return false;
			}
			members.add(index);
			indexToRowGroupName.put(index, rowGroupName);
		}

		this.indexToRowGroupName.putAll(indexToRowGroupName);
		rowGroup.members.addAll(members);
		notifyListeners();
		return true;
	}
	
    /**
     * Removes the rows from the group specified.
     */
    public void removeRowIndexesFromGroup(String rowGroupName, Integer... bodyRowIndexes) {

        RowGroup rowGroup = nameToRowGroup.get(rowGroupName);
        List<Integer> indexesToRemove = Arrays.asList(bodyRowIndexes);

        for (Integer rowIndex : indexesToRemove) {
            if (indexToRowGroupName.get(rowIndex).equals(rowGroupName)) {
                indexToRowGroupName.remove(rowIndex);
            }
        }

        rowGroup.getMembers().removeAll(indexesToRemove);
        notifyListeners();
    } 
	
	// Getters
	
	/*
	 * Do not expose private RowGroup class
	 */
	private RowGroup getRowGroupForIndex(int bodyRowIndex) {
		Integer key = Integer.valueOf(bodyRowIndex);
		if (indexToRowGroupName.containsKey(key)) {
			return nameToRowGroup.get(indexToRowGroupName.get(key));
		}
		return null;
	}
	
	public String getRowGroupNameForIndex(int bodyRowIndex) {
		if (isPartOfAGroup(bodyRowIndex)) {
			return getRowGroupForIndex(bodyRowIndex).name;
		}
		return null;
	}
	
	public List<String> getRowGroupNames() {
		final List<String> groupNames = new ArrayList<String>();
		groupNames.addAll( nameToRowGroup.keySet() );
		return Collections.unmodifiableList(groupNames);
	}
	
	public RowGroup getRowGroupForName(String rowGroupName) {
		return nameToRowGroup.get(rowGroupName);
	}
	
	/**
	 * @return Unmodifiable list of other row indexes in the same group as this index
	 */
	public List<Integer> getRowIndexesInGroup(int bodyRowIndex) {
		RowGroup group = getRowGroupForIndex(bodyRowIndex);
		if(group == null) {
			return Collections.emptyList();
		}
		return Collections.unmodifiableList(group.members);
	}
	
	/**
	 * Helper method to get the row positions for a specified layer
	 * 
	 * @param layer
	 * @param bodyRowIndexes
	 * @return
	 */
	public List<Integer> getRowPositionsInGroup(IUniqueIndexLayer layer, Collection<Integer> bodyRowIndexes) {
		final List<Integer> rowPositions = new ArrayList<Integer>();
		for( Integer bodyRowIndex : bodyRowIndexes ) {
			rowPositions.add( layer.getRowPositionByIndex(bodyRowIndex) );
		}
		return Collections.unmodifiableList(rowPositions);
	}
	
	/**
	 * @return Unmodifiable list of static row indexes (which remains visible
	 * when the group is collapsed) in the same group as this index
	 */
	public List<Integer> getStaticRowIndexesInGroup(int bodyRowIndex) {
	
		RowGroup group = getRowGroupForIndex(bodyRowIndex);
		if(group == null) {
			return Collections.emptyList();
		}
		return Collections.unmodifiableList(group.staticRowIndexes);
	}	
	
	public boolean isPartOfAGroup(int bodyRowIndex){
		return indexToRowGroupName.containsKey(Integer.valueOf(bodyRowIndex));
	}
	
	public void clear() {
		nameToRowGroup.clear();
		indexToRowGroupName.clear();
	}

	/**
	 * Number of rows in the Group which the bodyRowIndex belongs to.
	 */
	public int sizeOfGroup(int bodyRowIndex) {
		return getRowIndexesInGroup(bodyRowIndex).size();
	}
	
	/**
	 * Number of static rows in the Group which the bodyRowIndex belongs
	 * to.
	 */
	public int sizeOfStaticRows(int bodyRowIndex) {
		return getStaticRowIndexesInGroup(bodyRowIndex).size();
	}
	
	/**
	 * @return TRUE if no row groups exist
	 */
	public boolean isEmpty() {
		return nameToRowGroup.size() == 0;
	}
	
	/**
	 * @return TRUE if the row group this index belongs to is collapsed
	 */
	public boolean isCollapsed(int bodyRowIndex) {
		if (isPartOfAGroup(bodyRowIndex)) {
			return getRowGroupForIndex(bodyRowIndex).collapsed;
		}
		return false;
	}
	
	public boolean isCollapsed(String rowGroupName) {
		boolean collpased = false;
		if (nameToRowGroup.containsKey(rowGroupName)) {
			collpased = nameToRowGroup.get(rowGroupName).collapsed;
		}
		return collpased;
	}
	
	/**
	 * @return TRUE if the row group this index is collapseable
	 */
	public boolean isCollapseable(int bodyRowIndex) {
		if (isPartOfAGroup(bodyRowIndex)) {
			return getRowGroupForIndex(bodyRowIndex).collapseable;
		}
		return false;
	}

	public boolean isCollapseable(String rowGroupName) {
		boolean collpaseable = false;
		if (nameToRowGroup.containsKey(rowGroupName)) {
			collpaseable = nameToRowGroup.get(rowGroupName).collapseable;
		}
		return collpaseable;
	}
	
	/**
	 * @return TRUE if <code>bodyRowIndex</code> is contained in the list
	 * of static rows of the row group this index belongs to
	 */
	public boolean isStaticRow(int bodyRowIndex) {
		if (isPartOfAGroup(bodyRowIndex)) {
			return getRowGroupForIndex(bodyRowIndex).staticRowIndexes.contains(bodyRowIndex);
		}
		return false;
	}
	
	/**
	 * @return TRUE if a group by this name exists
	 */
	public boolean isAGroup(String cellValue) {
		return nameToRowGroup.containsKey(cellValue);
	}
	
	protected void collapse(int bodyRowIndex) {
		if (isPartOfAGroup(bodyRowIndex)) {
			getRowGroupForIndex(bodyRowIndex).collapsed = true;
		}
		notifyListeners();
	}

	protected void expand(int bodyRowIndex) {
		if (isPartOfAGroup(bodyRowIndex)) {
			getRowGroupForIndex(bodyRowIndex).collapsed = false;
		}
		notifyListeners();
	}
	
	/**
	 * Toggle the expand/collapse state of the Row Group
	 */
	public RowGroup toggleRowGroupExpandCollapse(int bodyRowIndex) {
		if (isPartOfAGroup(bodyRowIndex)) {
			RowGroup rowGroup = getRowGroupForIndex(bodyRowIndex);
			rowGroup.collapsed = !rowGroup.collapsed;
			notifyListeners();
			return rowGroup;
		}
		return null;
	}
	
	// Unbreakable Groups

	/**
	 * If a group is marked as unbreakable, the composition of the group cannot be changed.
	 *    Rows cannot be added or removed from the group.
	 *    Rows may be reorder within the group.
	 * @return TRUE if the operation was successful.
	 * @see NTBL 393
	 */
	public boolean setGroupUnBreakable(int bodyRowIndex) {
		if (isPartOfAGroup(bodyRowIndex)) {
			getRowGroupForIndex(bodyRowIndex).unbreakable = true;
			return true;
		}
		return false;
	}
	
	// *** Row Group ***

	public class RowGroup {
		
		/** Body row indexes */
		private final LinkedList<Integer> members = new LinkedList<Integer>();
		
		/** row indexes which remain visible when collapsing this group */
		private LinkedList<Integer> staticRowIndexes = new LinkedList<Integer>();
		
		final String name;
		
		boolean collapsed = false;
		boolean collapseable = true;
		
		public boolean unbreakable = false;

		RowGroup(String groupName) {
			this.name = groupName;
		}

		int getMemberCount(){
			return members.size();
		}

		public List<Integer> getMembers() {
			return members;
		}
		
		public List<Integer> getMembersSorted() {			
			final List<Integer> sortedMembers = new ArrayList<Integer>();
			sortedMembers.addAll( members );
			Collections.sort(sortedMembers);			
			return Collections.unmodifiableList( sortedMembers );
		}
		
		/**
		 * @return the row indexes which remains visible when collapsing
		 * this group
		 */
		public LinkedList<Integer> getStaticRowIndexes() {
			return staticRowIndexes;
		}
		
		/**
		 * @return the number of columns which remains visible when collapsing
		 * this group
		 */
		int getStaticRowIndexesCount() {
			return staticRowIndexes.size();
		}

		@Override
		public String toString() {
			return "Column Group:\n\t name: " + name
			+ "\n\t collapsed: " + collapsed
			+ "\n\t unbreakable: " + unbreakable
			+ "\n\t members: " + ObjectUtils.toString(members) + "\n"
			+ "\n\t staticColumns: " + ObjectUtils.toString(staticRowIndexes) + "\n";
		}
	}
}
